#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <windows.h>
#include <commctrl.h>
#include "widgets.h"

#define WIN_STYLE	WS_OVERLAPPEDWINDOW & ~WS_THICKFRAME

struct wgt_window {
	HWND handle;
	struct wgt_rect rect;

	HDC dc;
	char cname[32];
	struct wgt_widget *wlist;

	int close_quit;
	int (*cb_close)(struct wgt_window*);
};

struct wgt_widget {
	HWND handle;
	struct wgt_rect rect;

	struct wgt_window *win;
	struct wgt_widget *next;

	char **itemlist;
	int num_items;

	int disabled;

	wgt_callback cb_click;
	wgt_callback cb_modify;
};


static LRESULT WINAPI event_handler(HWND win, unsigned int msg, WPARAM wparam, LPARAM lparam);
static struct wgt_widget *find_widget(struct wgt_window *win, HWND handle);


void *wgt_window_handle(struct wgt_window *win)
{
	return win->handle;
}

void *wgt_widget_handle(struct wgt_widget *w)
{
	return w->handle;
}

unsigned int wgt_window_xid(struct wgt_window *win)
{
	return 0;
}

unsigned int wgt_widget_xid(struct wgt_widget *w)
{
	return 0;
}


struct wgt_window *wgt_window(const char *title, int width, int height)
{
	struct wgt_window *win;
	int tlen;
	WNDCLASS wc;
	HINSTANCE hinst;
	DWORD err;
	RECT rect;

	if(!(win = calloc(1, sizeof *win))) {
		fprintf(stderr, "wgl_create_window: failed to allocate window structure\n");
		return 0;
	}
	win->close_quit = 1;	/* by default quit when the window closes */

	tlen = strlen(title);
	if(tlen > sizeof win->cname - 10) {
		tlen = sizeof win->cname - 10;
	}
	strcpy(win->cname, "WGTCLASS-");
	memcpy(win->cname + 9, title, tlen);
	win->cname[tlen + 9] = 0;

	hinst = GetModuleHandle(0);

	memset(&wc, 0, sizeof wc);
	wc.hInstance = hinst;
	wc.hbrBackground = GetSysColorBrush(COLOR_3DFACE);
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hIcon = LoadIcon(0, IDI_APPLICATION);
	wc.lpszClassName = win->cname;
	wc.lpfnWndProc = event_handler;
	wc.cbWndExtra = 4;
	if(!RegisterClass(&wc)) {
		fprintf(stderr, "wgt_create_window: failed to register window class\n");
		return 0;
	}

	rect.left = rect.top = 0;
	rect.right = width;
	rect.bottom = height;
	AdjustWindowRect(&rect, WIN_STYLE, FALSE);

	if(!(win->handle = CreateWindow(win->cname, title, WIN_STYLE, CW_USEDEFAULT,
			CW_USEDEFAULT, rect.right - rect.left, rect.bottom - rect.top,
			0, 0, hinst, 0))) {
		fprintf(stderr, "wgt_create_window: failed to create window\n");
		UnregisterClass(title, hinst);
		return 0;
	}
	win->dc = GetDC(win->handle);
	SetLastError(0);
	if(!SetWindowLong(win->handle, GWL_USERDATA, (long)win) && (err = GetLastError())) {
		fprintf(stderr, "wgt_create_window: failed to store window user data (err: %d)\n", (int)err);
		UnregisterClass(title, hinst);
		DestroyWindow(win->handle);
		return 0;
	}
	ShowWindow(win->handle, 1);

	GetWindowRect(win->handle, &rect);

	win->rect.x = rect.left;
	win->rect.y = rect.top;
	win->rect.width = rect.right - rect.left;
	win->rect.height = rect.bottom - rect.top;

	return win;
}

void wgt_destroy_window(struct wgt_window *win)
{
	while(win->wlist) {
		struct wgt_widget *w = win->wlist;
		win->wlist = win->wlist->next;
		wgt_destroy_widget(w);
		free(w);
	}

	if(win->handle) {
		ReleaseDC(win->handle, win->dc);
		DestroyWindow(win->handle);
		UnregisterClass(win->cname, GetModuleHandle(0));
	}
}

void wgt_free_window(struct wgt_window *win)
{
	wgt_destroy_window(win);
	free(win);
}

void wgt_destroy_widget(struct wgt_widget *w)
{
	DestroyWindow(w->handle);
	free(w->itemlist);
}

void wgt_resize_window(struct wgt_window *win, int width, int height)
{
	RECT rect;

	win->rect.width = width;
	win->rect.height = height;

	rect.left = win->rect.x;
	rect.top = win->rect.y;
	rect.right = rect.left + width;
	rect.bottom = rect.top + height;
	if(AdjustWindowRect(&rect, WIN_STYLE, FALSE)) {
		width = rect.right - rect.left;
		height = rect.bottom - rect.top;
	}
	MoveWindow(win->handle, win->rect.x, win->rect.y, width, height, TRUE);
}

void wgt_quit_on_close(struct wgt_window *win, int quit)
{
	win->close_quit = quit;
}

void wgt_close_action(struct wgt_window *win, int (*func)(struct wgt_window*))
{
	win->cb_close = func;
}

void wgt_enable_widget(struct wgt_widget *w)
{
	w->disabled = 0;
	EnableWindow(w->handle, 1);
}

void wgt_disable_widget(struct wgt_widget *w)
{
	w->disabled = 1;
	EnableWindow(w->handle, 0);
}

int wgt_widget_enabled(struct wgt_widget *w)
{
	return !w->disabled;
}

struct wgt_window *wgt_widget_window(struct wgt_widget *w)
{
	return w->win;
}

struct wgt_rect *wgt_widget_rect(struct wgt_widget *w, struct wgt_rect *rect)
{
	if(rect) {
		*rect = w->rect;
	}
	return &w->rect;
}

void wgt_move_widget(struct wgt_widget *w, int x, int y)
{
	MoveWindow(w->handle, x, y, w->rect.width, w->rect.height, TRUE);
	w->rect.x = x;
	w->rect.y = y;
}

void wgt_resize_widget(struct wgt_widget *w, int width, int height)
{
	MoveWindow(w->handle, w->rect.x, w->rect.y, width, height, TRUE);
	w->rect.width = width;
	w->rect.height = height;
}

int wgt_xpos_after(struct wgt_widget *w, int pad)
{
	if(pad == WGT_AUTO) {
		pad = 8;
	}
	return w->rect.x + w->rect.width + pad;
}

int wgt_ypos_after(struct wgt_widget *w, int pad)
{
	if(pad == WGT_AUTO) {
		pad = 8;
	}
	return w->rect.y + w->rect.height + pad;
}

int wgt_string_size(struct wgt_window *win, const char *s, struct wgt_rect *rect)
{
	SIZE sz;

	GetTextExtentPoint32(win->dc, (char*)s, strlen(s), &sz);
	if(rect) {
		rect->x = rect->y = 0;
		rect->width = sz.cx;
		rect->height = sz.cy;
	}
	return sz.cx;
}


struct wgt_widget *wgt_label(struct wgt_window *win, const char *text, int x, int y)
{
	struct wgt_widget *w;
	struct wgt_rect textsz;

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_label: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	wgt_string_size(win, text, &textsz);

	if(!(w->handle = CreateWindow("STATIC", text, WS_CHILD | WS_VISIBLE | SS_SIMPLE,
			x, y, textsz.width, textsz.height, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_label: failed to create window\n");
		free(w);
		return 0;
	}

	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = textsz.width;
	w->rect.height = textsz.height;

	w->next = win->wlist;
	win->wlist = w;
	return w;
}

struct wgt_widget *wgt_button(struct wgt_window *win, const char *text, int x, int y,
			int width, int height, wgt_callback clickfunc)
{
	struct wgt_widget *w;
	struct wgt_rect textsz;

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_button: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	if(width < 0 || height < 0) {
		wgt_string_size(win, text, &textsz);
		if(width < 0) width = textsz.width * 5 / 3;
		if(height < 0) height = textsz.height * 10 / 7;
	}

	if(!(w->handle = CreateWindow("BUTTON", text, WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
			x, y, width, height, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_button: failed to create window\n");
		free(w);
		return 0;
	}

	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = width;
	w->rect.height = height;

	w->cb_click = clickfunc;
	w->next = win->wlist;
	win->wlist = w;
	return w;
}

struct wgt_widget *wgt_checkbox(struct wgt_window *win, const char *text, int on,
			int x, int y, int width, int height, wgt_callback modfunc)
{
	struct wgt_widget *w;
	struct wgt_rect textsz;
	int check_width;

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_checkbox: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	check_width = GetSystemMetrics(SM_CXMENUCHECK);

	if(width < 0 || height < 0) {
		wgt_string_size(win, text, &textsz);
		if(width < 0) width = textsz.width + check_width + 10;
		if(height < 0) height = textsz.height;
	}

	if(!(w->handle = CreateWindow("BUTTON", text, WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX,
			x, y, width, height, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_checkbox: failed to create window\n");
		free(w);
		return 0;
	}
	SendMessage(w->handle, BM_SETCHECK, on ? BST_CHECKED : BST_UNCHECKED, 0);

	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = width;
	w->rect.height = height;

	w->cb_click = modfunc;	/* BN_CLICKED is sent for checkbox state changes */
	w->next = win->wlist;
	win->wlist = w;
	return w;
}

static void combosize(struct wgt_window *win, HWND cbwnd, const char **items, int num_items,
		int width, int height, struct wgt_rect *sz)
{
	int i, max_width, sum_height;
	struct wgt_rect textsz;

	wgt_string_size(win, "00", &textsz);
	max_width = width < 0 && width != WGT_AUTO ? -width : textsz.width;

	/* first get the height of the editbox */
	sum_height = SendMessage(cbwnd, CB_GETITEMHEIGHT, -1, 0);
	sz->y = sum_height;
	/* then add the height of the dropdown list items */
	/* XXX I can't figure out why, but it's not sufficient to show all items
	 * without some more height, so let's double it...
	 */
	sum_height *= 2;
	sum_height += SendMessage(cbwnd, CB_GETITEMHEIGHT, 0, 0) * num_items;
	sum_height += GetSystemMetrics(SM_CYEDGE) * 2;

	if(width < 0) {
		int sbar_width;
		for(i=0; i<num_items; i++) {
			wgt_string_size(win, items[i], &textsz);
			if(textsz.width > max_width) max_width = textsz.width;
		}
		sbar_width = GetSystemMetrics(SM_CXVSCROLL);
		width = max_width + sbar_width + GetSystemMetrics(SM_CXEDGE) * 2;
	}
	if(height < 0) height = sum_height;

	sz->width = width;
	sz->height = height;
}

struct wgt_widget *wgt_combo(struct wgt_window *win, const char **items, int num_items,
			int sel, int x, int y, int width, int height, wgt_callback modfunc)
{
	struct wgt_widget *w;
	struct wgt_rect rect;

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_combo: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	if(!(w->handle = CreateWindow("COMBOBOX", items[0] ? items[0] : "",
			WS_CHILD | WS_VISIBLE | CBS_DROPDOWNLIST | CBS_HASSTRINGS,
			x, y, 16, 16, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_combo: failed to create window\n");
		free(w);
		return 0;
	}

	combosize(win, w->handle, items, num_items, width, height, &rect);
	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = rect.width;
	w->rect.height = rect.y;	/* single line height for layout purposes */
	MoveWindow(w->handle, x, y, rect.width, rect.height, FALSE);

	wgt_combo_setitems(w, items, num_items);

	if(sel < 0) sel = 0;
	if(sel >= num_items) sel = num_items - 1;
	if(sel >= 0) {
		SendMessage(w->handle, CB_SETCURSEL, sel, 0);
	}

	w->cb_modify = modfunc;
	w->next = win->wlist;
	win->wlist = w;
	return w;
}

struct wgt_widget *wgt_progbar(struct wgt_window *win, int x, int y, int width, int height, int p)
{
	struct wgt_widget *w;
	INITCOMMONCONTROLSEX commctrl;

	commctrl.dwSize = sizeof commctrl;
	commctrl.dwICC = ICC_PROGRESS_CLASS;
	InitCommonControlsEx(&commctrl);

	if(!(w = calloc(1, sizeof *w))) {
		fprintf(stderr, "wgt_progbar: failed to allocate widget structure\n");
		return 0;
	}
	w->win = win;

	if(height < 0) {
		height = GetSystemMetrics(SM_CYVSCROLL);
	}

	if(!(w->handle = CreateWindow(PROGRESS_CLASS, 0, WS_CHILD | WS_VISIBLE, x, y,
			width, height, win->handle, 0, GetModuleHandle(0), 0))) {
		fprintf(stderr, "wgt_progbar: failed to create window\n");
		free(w);
		return 0;
	}

	/*
	SendMessage(w->handle, PBM_SETRANGE, 0, MAKELPARAM(0, 100));
	SendMessage(w->handle, PBM_SETSTEP, 1, 0);
	*/
	SendMessage(w->handle, PBM_SETPOS, p, 0);

	w->rect.x = x;
	w->rect.y = y;
	w->rect.width = width;
	w->rect.height = height;

	w->next = win->wlist;
	win->wlist = w;
	return w;
}

void wgt_set_text(struct wgt_widget *w, const char *text)
{
	SendMessage(w->handle, WM_SETTEXT, 0, (long)text);

	w->rect.width = wgt_string_size(w->win, text, 0);
	MoveWindow(w->handle, w->rect.x, w->rect.y, w->rect.width, w->rect.height, TRUE);
}

void wgt_checkbox_check(struct wgt_widget *w)
{
	SendMessage(w->handle, BM_SETCHECK, BST_CHECKED, 0);
}

void wgt_checkbox_uncheck(struct wgt_widget *w)
{
	SendMessage(w->handle, BM_SETCHECK, BST_UNCHECKED, 0);
}

int wgt_checkbox_checked(struct wgt_widget *w)
{
	return SendMessage(w->handle, BM_GETCHECK, 0, 0) == BST_CHECKED;
}

int wgt_combo_setitems(struct wgt_widget *w, const char **items, int num_items)
{
	int i;
	char **newit;
	struct wgt_rect size;

	if(!(newit = malloc(num_items * sizeof *newit))) {
		fprintf(stderr, "wgt_combo_setitems: failed to allocate item list\n");
		return -1;
	}
	for(i=0; i<num_items; i++) {
		if(!(newit[i] = strdup(items[i]))) {
			fprintf(stderr, "wgt_combo_setitems: failed to allocate item\n");
			while(--i >= 0) {
				free(newit[i]);
			}
			free(newit);
			return -1;
		}
	}

	for(i=0; i<w->num_items; i++) {
		free(w->itemlist[i]);
	}
	free(w->itemlist);
	SendMessage(w->handle, CB_RESETCONTENT, 0, 0);

	w->itemlist = newit;
	w->num_items = num_items;

	for(i=0; i<num_items; i++) {
		SendMessage(w->handle, CB_ADDSTRING, 0, (long)w->itemlist[i]);
	}

	combosize(w->win, w->handle, items, num_items, w->rect.width, WGT_AUTO, &size);
	MoveWindow(w->handle, w->rect.x, w->rect.y, w->rect.width, size.height, TRUE);

	SendMessage(w->handle, CB_SETCURSEL, 0, 0);
	return 0;
}

int wgt_combo_selected(struct wgt_widget *w)
{
	int sel = SendMessage(w->handle, CB_GETCURSEL, 0, 0);
	if(sel == CB_ERR) return -1;
	return sel;
}

const char *wgt_get_combo_item(struct wgt_widget *w, int idx)
{
	if(idx < 0 || idx >= w->num_items) {
		return 0;
	}
	return w->itemlist[idx];
}

void wgt_set_progress(struct wgt_widget *w, int p)
{
	SendMessage(w->handle, PBM_SETPOS, p, 0);
}

static LRESULT WINAPI event_handler(HWND wnd, unsigned int msg, WPARAM wparam, LPARAM lparam)
{
	int ncode;
	struct wgt_widget *w;
	struct wgt_window *win;

	win = (struct wgt_window*)GetWindowLong(wnd, GWL_USERDATA);

	switch(msg) {
	case WM_CLOSE:
		if(!win->cb_close || win->cb_close(win)) {
			if(win) {
				int quit = win->close_quit;
				wgt_destroy_window(win);
				memset(win, 0, sizeof *win);
				win->close_quit = quit;
			} else {
				DestroyWindow(wnd);
			}
		}
		break;

	case WM_DESTROY:
		if(win && win->close_quit) {
			PostQuitMessage(0);
		}
		break;

	case WM_COMMAND:
		if(!win || !(w = find_widget(win, (HWND)lparam))) {
			break;
		}
		ncode = HIWORD(wparam);

		switch(ncode) {
		case BN_CLICKED:
			if(w->cb_click) {
				w->cb_click(w);
			}
			break;

		case CBN_SELCHANGE:
			if(w->cb_modify) {
				w->cb_modify(w);
			}
			break;

		default:
			break;
		}

	default:
		return DefWindowProc(wnd, msg, wparam, lparam);
	}
	return 0;
}

static struct wgt_widget *find_widget(struct wgt_window *win, HWND handle)
{
	struct wgt_widget *w;

	w = win->wlist;
	while(w) {
		if(w->handle == handle) return w;
		w = w->next;
	}
	return 0;
}
